from ..phase import StopProgram
from ..tools import instrument_definition


@instrument_definition
def early_stop(ov, key, n, task=None, signal=StopProgram):
    called = False

    def _stop(value):
        # The stop signal may have the unfortunate effect of creating
        # another event, so this may get called twice.
        nonlocal called
        if not called:
            called = True
            if isinstance(signal, str):
                ov.log({"$event": signal})
            else:
                raise signal(value)

    yield ov.phases.init

    stream = ov.given
    if task is not None:
        stream = stream.where(task=task)
    stream = stream.where(key)
    stream.map_indexed(
        lambda _, idx: {"task": "early_stop", "progress": (idx + 1, n)}
    ).give()
    stream.skip(n) >> _stop



import glob
import os
import traceback

from ...errors import NotAvailable
from ...tools import instrument_definition
from ..utils import Monitor as Monitor2


def find_monitors():
    """Look for device monitor implementation (AMD or ROCm)"""
    backends = {}
    base = __file__
    base_module = "voir.instruments.gpu"
    module_path = os.path.dirname(os.path.abspath(base))
    pattern = os.path.join(module_path, "[A-Za-z]*")

    for module_path in glob.glob(pattern, recursive=False):
        module_file = module_path.split(os.sep)[-1]

        if module_file == "__init__.py":
            continue

        module_name = module_file.split(".py")[0]

        try:
            module = __import__(".".join([base_module, module_name]), fromlist=[""])
        except ImportError:
            print(traceback.format_exc())
            continue

        backends[module_name] = module

    return backends


BACKENDS = find_monitors()
DEVICESMI = None


def get_backends():
    global BACKENDS
    return BACKENDS.keys()


def _is_backend_available(backend):
    try:
        smi = backend.DeviceSMI()

        if len(smi.get_gpus_info()) > 0:
            return True

    except NotAvailable:
        return False


def deduce_backend():
    suitable = []
    for k, backend in BACKENDS.items():
        if backend.is_installed() and _is_backend_available(backend):
            suitable.append(k)

    if len(suitable) > 1:
        options = ", ".join(get_backends())
        raise Exception(
            f"Milabench found multiple vendors ({suitable}) and does not "
            f"know which kind to use. Please set $MILABENCH_GPU_ARCH to one of {options}."
        )

    elif len(suitable) == 0:
        return "cpu"

    return suitable[0]


def select_backend(arch=None):
    global DEVICESMI

    if DEVICESMI is not None:
        if DEVICESMI.arch == arch:
            return DEVICESMI
        DEVICESMI.close()
        DEVICESMI = None

    if arch is None:
        arch = deduce_backend()

    backend = BACKENDS.get(arch)

    if backend is not None and backend.is_installed():
        DEVICESMI = backend.DeviceSMI()
    else:
        raise NotAvailable(f"{arch} is not installed")

    return DEVICESMI


def gpu_info(smi):
    return {
        "arch": smi.arch,
        "gpus": smi.get_gpus_info(),
    }


def get_gpu_info(arch=None):
    return gpu_info(select_backend(arch))


def _visible_devices(smi):
    visible = smi.visible_devices

    if visible:
        ours = visible.split(",")
    else:
        ours = [str(x) for x in range(100)]

    return ours


@instrument_definition
def gpu_monitor(ov, poll_interval=10, arch=None):
    yield ov.phases.load_script

    smi = select_backend(arch)
    ours = _visible_devices(smi)

    def monitor():
        data = {
            gpu["device"]: {
                "memory": [
                    gpu["memory"]["used"],
                    gpu["memory"]["total"],
                ],
                "load": gpu["utilization"]["compute"],
                "temperature": gpu["temperature"],
            }
            for gpu in smi.get_gpus_info().values()
            if str(gpu["device"]) in ours
        }
        ov.give(task="main", gpudata=data)

    monitor_thread = Monitor2(poll_interval, monitor)
    monitor_thread.start()
    try:
        yield ov.phases.run_script
    finally:
        monitor_thread.stop()
        monitor()



import re
from collections import deque

oldhex = "0123456789abcdef"
newhex = "0123456789:;<=>?"

hexsub = dict(zip(oldhex, newhex))
hexsub_back = dict(zip(newhex, oldhex))


def encode_control(s):
    encoded = "".join([hexsub[c] for c in s.encode("utf8").hex()])
    return f"\033[{encoded}z"


def decode_control(s):
    s = s[2:-1]
    decoded_hex = "".join([hexsub_back[c] for c in s])
    return bytes.fromhex(decoded_hex).decode("utf8")


def decode_control_stream(s):
    pass


class SmuggleWriter:
    def __init__(self, stream):
        self.stream = stream

    def __enter__(self):
        self.stream.__enter__()
        return self

    def __exit__(self, typ=None, exc=None, tb=None):
        self.stream.__exit__(typ, exc, tb)
        return self

    def write(self, txt):
        return self.stream.write(encode_control(txt))

    def flush(self):
        return self.stream.flush()

    def close(self):
        return self.stream.close()


class LineAccumulator:
    def __init__(self):
        self.lines = deque()
        self.current = ""

    def process(self, char):
        self.current += char
        if char == "\n":
            self.lines.append(self.current)
            self.current = ""


class Decoder:
    def __init__(self, principal):
        self.principal = principal
        self.out = LineAccumulator()
        self.data = LineAccumulator()
        self.code = ""
        self.coding = False

    def endcode(self):
        self.coding = False
        if re.match(string=self.code, pattern="\033\\[[0-9:;<=>?]*z"):
            for char in decode_control(self.code):
                self.data.process(char)
        else:
            self.out.current += self.code
        self.code = ""

    def process_char(self, char):
        if self.coding:
            if ord(char) < 0x20:
                self.out.process(char)
                self.endcode()
            elif ord(char) >= 0x40 and char != "[":
                self.code += char
                self.endcode()
            else:
                self.code += char

        elif char == "\033":
            self.coding = True
            self.code += char

        else:
            self.out.process(char)

    def getline(self, which):
        if which == "out":
            if self.out.lines:
                return self.out.lines.popleft()
        elif which == "data":
            if self.data.lines:
                return self.data.lines.popleft()
        return None

    def readline(self, which):
        while (result := self.getline(which)) is None:
            nxt = self.principal.read(1)
            if not nxt:
                return None
            self.process_char(nxt.decode("utf8"))
        return result


class MultimodalFile:
    def __init__(self, decoder, which, name):
        self.name = name
        self.decoder = decoder
        self.which = which

    def fileno(self):
        return self.name

    def readline(self):
        zazz = self.decoder.readline(self.which)
        # print(zazz, self.decoder.principal)
        return zazz



import time
from threading import Thread


class Monitor(Thread):
    def __init__(self, delay, func):
        super().__init__(daemon=True)
        self.stopped = False
        self.delay = delay
        self.func = func

    def run(self):
        while not self.stopped:
            time.sleep(self.delay)
            self.func()

    def stop(self):
        self.stopped = True



from contextvars import ContextVar

from giving import give

current_overseer = ContextVar("current_overseer", default=None)


def log(**kwargs):
    ov = current_overseer.get()
    if ov is not None:
        ov.log(kwargs)


def iterate(
    task: str, iterable, report_batch=False, ignore_loading=False, batch_size=None
):
    assert isinstance(task, str)
    try:
        n = len(iterable)
    except TypeError:
        n = None

    def prog(i):
        if n is not None:
            give(progress=(i, n))

    i = 0
    with give.inherit(task=task):
        prog(0)
        it = iter(iterable)
        while True:
            if i == n:
                break
            i += 1

            def get_batch():
                batch = next(it)
                if batch_size is None:
                    kwargs = {"batch": batch}
                elif callable(batch_size):
                    kwargs = {"batch_size": batch_size(batch)}
                else:
                    kwargs = {"batch_size": batch_size}
                return batch, kwargs

            try:
                if not report_batch:
                    batch, kwargs = get_batch()
                    yield batch
                elif ignore_loading:
                    batch, kwargs = get_batch()
                    with give.wrap("step", **kwargs):
                        yield batch
                else:
                    empty_kwargs = (
                        {"batch": None} if batch_size is None else {"batch_size": None}
                    )
                    with give.wrap("step", **empty_kwargs) as extra:
                        batch, kwargs = get_batch()
                        extra.update(kwargs)
                        yield batch
            except StopIteration:
                break
            prog(i)



import time
from typing import Union

from giving import Given
from ovld import ovld

from ..tools import instrument_definition


@ovld
def _parse_duration(x: Union[int, float]):
    return (False, x)


@ovld
def _parse_duration(x: str):  # noqa: F811
    if x.endswith("s"):
        return (True, float(x[:-1]))
    else:
        return (False, float(x))


def default_batch_size_calc(batch):
    if isinstance(batch, (list, tuple)):
        return len(batch[0])
    else:
        return len(batch)


@instrument_definition
def rate(
    ov,
    interval=1,
    skip=0,
    method=None,
    batch_size_calc=default_batch_size_calc,
    sync=None,
):
    """Compute a rate of computation, in items/s

    The ``rate`` instrument will be triggered by the following calls:

    .. code-block::

        # Method 1: "delta"
        give(task="task_name", batch_size=64)
        give(task="task_name", batch=np.ones((64, 1024, 1024)))
        give(task="task_name", batch=[np.ones((64, 1024, 1024), np.ones((64,))])

        # Method 2: "wrap"
        with giving.wrap("step", task="task_name", batch_size=64):
            ...

        with giving.wrap("step", task="task_name", batch=np.ones((64, 1024, 1024))):
            ...

        with giving.wrap("step", task="task_name", batch_size=None) as data:
            ...
            data["batch_size"] = 64

    In all examples, the batch size will be calculated to be 64. "task_name" is usually
    "train" or "eval".

    In Method 1, the batch size will be divided by the time between two calls to ``give``
    with the same task, yielding the number of items processed by second.

    In Method 2, the batch size will be divided by the time spent inside the ``with``
    statement, again yielding the number of items processed by second.

    See ``voir.iterate`` for an easy interface to this feature:

    .. code-block::

        for x in voir.iterate("train", loader, report_batch=True):
            ...

    Iterate will wrap every iteration of the loop with ``with giving.wrap("step", ...)``.

    Arguments:
        ov: The overseer.
        interval: Either the number of step events between two rate calculations, as an
            int, or the string "Ns" where N is the number of seconds between two rate
            calculations.
        skip: The number of rate calculations to skip (think of it as a warmup period).
        method: Either "delta" or "wrap" depending on whether you want to use Method
            1 or Method 2 as described above.
        batch_size_calc: A function to compute the batch size from the "batch" (e.g.
            ``batch_size_calc=len``). By default, if the batch is a list or a tuple,
            we will assume that it is probably an ``(input, target)`` tuple or something
            of the sort and we take the len of the first element.
        sync: A function to call whenever we want to calculate the rate and which will
            be timed along with the rest. For example, if running on a CUDA GPU, pass
            ``torch.cuda.synchronize`` to ensure that all pending calculations are taken
            into account in the rate calculation.
    """

    yield ov.phases.load_script

    interval_is_time, interval = _parse_duration(interval)

    # Build stream of task/time/batch_size
    if method is None or method == "delta":
        steps_w_batch = ov.given.where("task", "batch", "!batch_size", "!$wrap").kmap(
            task=lambda task: task, batch_size=lambda batch: batch_size_calc(batch)
        )
        steps_w_batch_size = ov.given.where("task", "batch_size", "!$wrap")
        times = times_step = (
            (steps_w_batch | steps_w_batch_size)
            .augment(time=lambda: time.time_ns())
            .pairwise()
            .starmap(
                lambda x, y: {
                    "task": y["task"],
                    "time": (y["time"] - x["time"]) / 1_000_000_000,
                    "batch_size": y["batch_size"],
                }
            )
        )

    if method is None or method == "wrap":

        def _timewrap():
            t0 = time.time_ns()
            results = yield
            t1 = time.time_ns()
            task = results["task"]
            if "batch_size" in results:
                bs = results["batch_size"]
                if bs is None:
                    return None
                seconds = (t1 - t0) / 1_000_000_000
                return {
                    "task": task,
                    "time": seconds,
                    "batch_size": bs,
                }
            elif "batch" in results:
                data = results["batch"]
                if data is None:
                    return None
                seconds = (t1 - t0) / 1_000_000_000
                return {
                    "task": task,
                    "time": seconds,
                    "batch_size": batch_size_calc(data),
                }
            else:
                return None

        times = times_wrap = ov.given.wmap("step", _timewrap).filter(lambda x: x)

    if method is None:
        times = times_step | times_wrap

    grouped_by_task = times.group_by(lambda data: data["task"])

    @grouped_by_task.subscribe
    def setup_pipeline(times):
        times = Given(_obs=times)

        # Group by interval
        if interval_is_time:
            times = times.buffer_with_time(interval)
        else:
            times = times.buffer_with_count(interval)

        if skip:
            times = times.skip(skip)

        # Compute the final metric
        @times.subscribe
        def _(elems):
            t = 0
            if sync is not None:
                t0 = time.time_ns()
                sync()
                t1 = time.time_ns()
                t += (t1 - t0) / 1_000_000_000

            t += sum(e["time"] for e in elems)
            n = sum(e["batch_size"] for e in elems)

            if n and t:
                ov.give(rate=n / t, units="items/s", task=elems[0]["task"])



from collections import Counter, defaultdict


class Plain:
    def __init__(self, x, fmt="{}"):
        self._object = x
        self.fmt = fmt

    def __rich__(self):
        return self.fmt.format(str(self._object))


def make_table(rows):
    from rich.table import Table

    table = Table.grid(padding=(0, 3, 0, 0))
    table.add_column("task", style="bold yellow")
    table.add_column("key", style="bold green")
    table.add_column("value")

    for task, values in rows.items():
        values = dict(values)
        progress = values.pop("progress", None)
        if progress is not None:
            table.add_row(task, "", progress)
            task = ""
        for key, value in values.items():
            table.add_row(task, key, value)
            task = ""  # Avoid displaying the task for the other rows

    return table


def dash(ov):
    """Create a simple terminal dashboard using rich.

    This displays a live table of the last value for everything given.
    """
    from rich.console import Console, Group
    from rich.live import Live
    from rich.panel import Panel
    from rich.pretty import Pretty
    from rich.progress import (
        BarColumn,
        Progress,
        TextColumn,
        TimeRemainingColumn,
    )

    yield ov.phases.init

    # Current rows are stored here
    rows = defaultdict(dict)

    console = Console()
    event_stats = Counter()
    event_types = {}

    def statistics(data):
        event_stats.update(list(data))
        event_types.update({k: type(v) for k, v in data.items()})

    # This updates the table every time we get new values
    def update_with(values):
        values = dict(values)

        units = values.pop("units", None)
        task = values.pop("task", "default")
        rowgroup = rows[task]

        if (
            isinstance(progress := values.get("progress", None), (list, tuple))
            and len(progress) == 2
        ):
            k = "progress"
            completed, total = progress
            if k not in rowgroup:
                progress_bar = Progress(
                    BarColumn(),
                    TimeRemainingColumn(),
                    TextColumn("({task.completed}/{task.total})"),
                )
                progress_bar._task = progress_bar.add_task(k)
                rowgroup[k] = progress_bar
            progress_bar = rowgroup[k]
            progress_bar.update(progress_bar._task, completed=completed, total=total)
            return

        for k, v in values.items():
            if k.startswith("$"):
                continue
            if k in rowgroup:
                rowgroup[k]._object = v
            else:
                if units:
                    rowgroup[k] = Plain(v, f"{{}} {units}")
                else:
                    rowgroup[k] = Pretty(v)

        if event_stats:
            strings = [
                f"{n}x {k} ({event_types[k].__name__})" for k, n in event_stats.items()
            ]
            lv.update(Panel(Group(make_table(rows), ", ".join(strings))))
        else:
            lv.update(Panel(make_table(rows)))

    panel = Panel("")

    ov.log >> update_with
    ov.given >> statistics

    with Live(panel, refresh_per_second=4, console=console) as lv:
        yield ov.phases.finalize(-1000)



import ast
import inspect
import tokenize
import warnings
from argparse import ArgumentParser
from dataclasses import MISSING, dataclass, fields, is_dataclass
from functools import partial
from textwrap import dedent
from typing import Union

from omegaconf import OmegaConf
from ovld import meta, ovld


@dataclass
class Info:
    name: str
    type: type
    help: str
    annotation: object
    prefix: str
    parser: ArgumentParser
    group: ArgumentParser


def _dash(base):
    base = base.replace("_", "-")
    if len(base) == 1:
        return f"-{base}"
    else:
        return f"--{base}"


@ovld
def contribute(default: meta(is_dataclass), info: Info):
    group = info.parser
    hlp = getattr(info.type, "__help__", None) or info.type.__name__
    if hlp:
        group = info.parser.add_argument_group(hlp)
    docs = get_attribute_docstrings(info.type)
    if info.prefix is None:
        pfx = ""
    else:
        pfx = f"{info.prefix}{info.name}."
    info.parser.constructors[pfx] = (info.type, f"{info.prefix or ''}{info.name}")
    for field in fields(info.type):
        contribute[field.type, Info](
            field.default if default is MISSING else getattr(default, field.name),
            Info(
                name=field.name,
                type=field.type,
                help=docs.get(field.name, None),
                annotation=None,
                prefix=pfx,
                parser=info.parser,
                group=group,
            ),
        )


@ovld
def contribute(default: bool, info: Info):  # noqa: F811
    hlp = info.help
    pth = f"{info.prefix}{info.name}"

    is_default = "(Default) " if default is True else ""
    info.group.add_argument(
        # f"--{pth}",
        _dash(pth),
        action="store_true",
        dest=pth,
        default=default,
        help=f"{is_default}{hlp}",
    )

    pth_no = f"{info.prefix}no-{info.name}"
    is_default = "(Default) " if default is False else ""
    info.group.add_argument(
        _dash(pth_no),
        action="store_false",
        dest=pth,
        default=default,
        help=hlp and f"{is_default}Do not {hlp[0].lower()}{hlp[1:]}",
    )


@ovld
def contribute(default: Union[int, float, str], info: Info):  # noqa: F811
    info.group.add_argument(
        _dash(f"{info.prefix}{info.name}"),
        type=info.type,
        default=None if default is MISSING else default,
        required=default is MISSING,
        metavar=info.name.upper(),
        help=info.help,
    )


def _expand(args, constructors):
    for name in sorted(constructors, key=len, reverse=True):
        ln = len(name)
        cons, dest = constructors[name]
        cons_args = {}
        field_names = {f.name for f in fields(cons)}
        for k, v in list(vars(args).items()):
            if k.startswith(name):
                field_name = k[ln:]
                if field_name in field_names:
                    cons_args[field_name] = v
                    delattr(args, k)
        obj = cons(**cons_args)
        setattr(args, dest, obj)
    return args


class ExtendedArgumentParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.constructors = {}
        self.base_configs = {}
        self.base_configs_locked = False
        self.used_base_configs = set()

    def merge_base_config(self, config):
        if self.base_configs_locked:
            raise Exception(
                "Cannot merge a base config after add_from_model is called."
            )
        self.base_configs = OmegaConf.merge(self.base_configs, config)

    def add_from_model(self, dest, model, flatten=True):
        self.base_configs_locked = True

        if isinstance(model, type):
            typ = model
            model = MISSING
        else:
            typ = type(model)

        self.used_base_configs.add(dest)
        if dest in self.base_configs:
            assert model is MISSING
            model = OmegaConf.merge(OmegaConf.structured(typ), self.base_configs[dest])

        contribute[typ, Info](
            model,
            Info(
                name=dest,
                type=typ,
                help=None,
                annotation=None,
                prefix=None if flatten else dest,
                parser=self,
                group=self,
            ),
        )

    def _parse_known_args(self, *args, **kwargs):
        unused = set(self.base_configs) - self.used_base_configs
        if unused:
            warnings.warn(
                f"Configuration blocks {unused} were not used. Valid blocks are: {self.used_base_configs}. Did you forget a nesting level?"
            )
        ns, args = super()._parse_known_args(*args, **kwargs)
        _expand(ns, self.constructors)
        return ns, args


#############################
# Extracting the docstrings #
#############################


def scrape_comments(src):
    lines = bytes(src, encoding="utf8").splitlines(keepends=True)
    return [
        (*tok.start, "COMMENT", tok.string[1:].strip())
        for tok in tokenize.tokenize(partial(next, iter(lines)))
        if tok.type == tokenize.COMMENT
    ]


class AttributeVisitor(ast.NodeVisitor):
    def __init__(self):
        self.data = []
        self.prefix = None

    def add_data(self, node, kind, content):
        self.data.append((node.lineno, node.col_offset, kind, content))

    def visit_body(self, name, stmts):
        old_prefix = self.prefix
        if self.prefix is None:
            self.prefix = ""
        else:
            self.prefix += f"{name}."
        for stmt in stmts:
            if (
                isinstance(stmt, ast.Expr)
                and isinstance(stmt.value, ast.Constant)
                and isinstance(stmt.value.value, str)
            ):
                self.add_data(stmt, "DOC", stmt.value.value)
            else:
                self.visit(stmt)
        self.prefix = old_prefix

    def visit_ClassDef(self, node):
        if self.prefix is not None:
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.name}")
        self.visit_body(node.name, node.body)

    def visit_FunctionDef(self, node):
        if self.prefix is not None:
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.name}")
        self.visit_body(node.name, node.body)

    def visit_Assign(self, node):
        self.generic_visit(node, may_assign=True)

    def visit_AnnAssign(self, node):
        self.generic_visit(node, may_assign=True)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.id}")

    def generic_visit(self, node, may_assign=False):
        if isinstance(node, ast.stmt) and not may_assign:
            self.add_data(node, "OTHER", None)
        super().generic_visit(node)


def scrape_docstrings(src):
    visitor = AttributeVisitor()
    visitor.visit(ast.parse(src))
    return visitor.data


def get_attribute_docstrings(cls):
    docs = {}
    current = None
    current_line = None
    for_next = []
    src = dedent(inspect.getsource(cls))
    data = scrape_comments(src) + scrape_docstrings(src)
    for line, _, kind, content in sorted(data):
        if kind == "COMMENT":
            if current is not None and current_line == line:
                docs[current].append(content)
            else:
                for_next.append(content)
        elif kind == "DOC" and current:
            docs[current].append(content)
        elif kind == "VARIABLE":
            docs[content] = for_next
            for_next = []
            current = content
            current_line = line
        elif kind == "OTHER":
            current = current_line = None
            for_next = []
    return {k: "\n".join(lines) for k, lines in docs.items()}



import importlib
import json
import os
import pkgutil
import sys
import traceback
from argparse import REMAINDER
from importlib.machinery import ModuleSpec
from pathlib import Path
from types import ModuleType

import yaml
from giving import SourceProxy
from ptera import probing, select

from voir.smuggle import SmuggleWriter

from .argparse_ext import ExtendedArgumentParser
from .helpers import current_overseer
from .phase import GivenPhaseRunner
from .scriptutils import split_script


class GiveToFile:
    def __init__(self, filename, fields=None, require_writable=True):
        self.fields = fields
        self.filename = filename
        if self.filename == 1:
            self.out = SmuggleWriter(sys.stdout)
        elif self.filename == 2:
            self.out = SmuggleWriter(sys.stderr)
        else:
            try:
                self.out = open(self.filename, "w", buffering=1)
            except OSError:
                if require_writable:
                    raise
                self.out = open(os.devnull, "w")
        self.out.__enter__()

    def log(self, data):
        try:
            txt = json.dumps(data)
        except TypeError:
            try:
                txt = json.dumps({"$unserializable": str(data)})
            except Exception:
                txt = json.dumps({"$unrepresentable": None})
        self.out.write(f"{txt}\n")

    def close(self):
        self.out.__exit__()


class LogStream(SourceProxy):
    def __call__(self, data):
        self._push(data)


class ProbeInstrument:
    def __init__(self, selector):
        self.selector = selector
        self.probe = self.__state__ = probing(self.selector)

    def __call__(self, ov):
        yield ov.phases.load_script(priority=0)
        with self.probe:
            yield ov.phases.run_script(priority=0)


class Overseer(GivenPhaseRunner):
    def __init__(self, instruments, logfile=None):
        self.argparser = ExtendedArgumentParser()
        self.argparser.add_argument("SCRIPT", nargs="?", help="The script to run")
        self.argparser.add_argument(
            "ARGV", nargs=REMAINDER, help="Arguments to the script"
        )
        self.argparser.add_argument(
            "-m",
            dest="MODULE",
            nargs=REMAINDER,
            help="Module or module:function to run",
        )

        super().__init__(
            phase_names=["init", "parse_args", "load_script", "run_script", "finalize"],
            args=(self,),
            kwargs={},
        )
        for instrument in instruments:
            self.require(instrument)
        self.logfile = logfile

    def on_overseer_error(self, e):
        self.log(
            {
                "$event": "overseer_error",
                "$data": {"type": type(e).__name__, "message": str(e)},
            }
        )
        print("=" * 80, file=sys.stderr)
        print(
            "voir: An error occurred in an overseer. Execution proceeds as normal.",
            file=sys.stderr,
        )
        print("=" * 80, file=sys.stderr)
        traceback.print_exception(type(e), e, e.__traceback__)
        print("=" * 80, file=sys.stderr)
        super().on_overseer_error(e)

    def probe(self, selector):
        return self.require(ProbeInstrument(select(selector, skip_frames=1)))

    def run_phase(self, phase):
        self.log({"$event": "phase", "$data": {"name": phase.name}})
        return super().run_phase(phase)

    def run(self, argv):
        self.log = LogStream()
        self.given.where("$event") >> self.log
        if self.logfile is not None:
            self.gtf = GiveToFile(self.logfile, require_writable=False)
            self.log >> self.gtf.log
        else:
            self.gtf = None

        with self.run_phase(self.phases.init):
            tmp_argparser = ExtendedArgumentParser(add_help=False)
            tmp_argparser.add_argument("--config", action="append", default=[])
            tmp_options, argv = tmp_argparser.parse_known_args(argv)
            for config in tmp_options.config:
                self.argparser.merge_base_config(yaml.safe_load(open(config, "r")))

        with self.run_phase(self.phases.parse_args):
            self.options = self.argparser.parse_args(argv)
            del self.argparser

        with self.run_phase(self.phases.load_script):
            script, argv, func = find_function(self.options)

        with self.run_phase(self.phases.run_script) as set_value:
            sys.argv = [script, *argv]
            set_value(func())

    def __call__(self, *args, **kwargs):
        token = current_overseer.set(self)
        try:
            super().__call__(*args, **kwargs)
        except BaseException as e:
            self.log(
                {
                    "$event": "error",
                    "$data": {"type": type(e).__name__, "message": str(e)},
                }
            )
            raise
        finally:
            with self.run_phase(self.phases.finalize):
                pass
            if self.gtf:
                self.gtf.close()
            current_overseer.reset(token)


def find_function(options):
    if script := options.SCRIPT:
        return script, options.ARGV, find_script(script)
    elif module_args := options.MODULE:
        module_spec, *argv = module_args
        if ":" in module_spec:
            module_name, field = module_spec.split(":", 1)
            module = importlib.import_module(module_name)
            return module_spec, argv, getattr(module, field)
        else:
            module_name = module_spec
            script = Path(pkgutil.get_loader(module_name).get_filename())
            if script.name == "__init__.py":
                script = script.parent / "__main__.py"
                module_name = f"{module_name}.__main__"
            script = str(script)
            return script, argv, find_script(script, module_name=module_name)
    else:
        sys.exit("Either SCRIPT or -m MODULE must be given.")


def find_script(script, module_name=None):
    prep, mainsection = split_script(script)
    mod = ModuleType("__main__")
    glb = vars(mod)
    glb["__file__"] = script
    if module_name:
        glb["__spec__"] = ModuleSpec(name=module_name, loader=None)
    sys.modules["__main__"] = mod
    exec(prep, glb, glb)
    return lambda: exec(mainsection, glb, glb)



import ast
import io


def split_script(script):
    """Split code that comes after all function definitions.

    Essentially, we want to be able to instrument functions in the main script, which
    requires evaluating the functions, but we want to do this before executing the main
    code. So we split off code that comes after function definitions so that we can evaluate
    the module and then evaluate that code separately.

    Code between function definitions will be evaluated right away, but the bulk usually
    comes after these definitions (because they need to use them).
    """

    with io.open_code(script) as f:
        source_code = f.read()

    tree = ast.parse(source_code, mode="exec")

    last_def = 0
    for i, stmt in enumerate(tree.body):
        if isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            last_def = i + 1

    mod_before = ast.copy_location(
        ast.Module(
            body=tree.body[:last_def],
            type_ignores=[],
        ),
        tree,
    )

    mod_after = ast.copy_location(
        ast.Module(
            body=tree.body[last_def:],
            type_ignores=[],
        ),
        tree,
    )

    return (
        compile(mod_before, script, "exec"),
        compile(mod_after, script, "exec"),
    )



import fnmatch

from ..tools import instrument_definition


def _keep(patterns, context):
    def operation(data):
        result = {}
        ok = False
        for k, v in data.items():
            if k in context or any(fnmatch.fnmatch(k, p) for p in context):
                result[k] = v
            if k in patterns or any(fnmatch.fnmatch(k, p) for p in patterns):
                result[k] = v
                ok = True
        return ok and result

    return operation


@instrument_definition
def log(ov, *patterns, context=[]):
    if not isinstance(context, (list, tuple)):
        context = [context]

    yield ov.phases.init

    more_context = {p[1:] for p in patterns if p.startswith("+")}
    context = {*more_context, *context}
    patterns = {p for p in patterns if not p.startswith("+")}

    ov.given.map(_keep(patterns, context)).filter(lambda x: x) >> ov.log



def is_installed():
    return True


class DeviceSMI:
    def get_gpus_info(self):
        return {}

    @property
    def arch(self):
        return "cpu"

    @property
    def visible_devices():
        return ""

    def close(self):
        pass



import os

from ...errors import NotAvailable

IMPORT_ERROR = None
try:
    from pynvml import nvmlInit
    from pynvml.nvml import (
        NVMLError_DriverNotLoaded,
        NVMLError_LibraryNotFound,
    )
    from pynvml.smi import nvidia_smi
except ImportError as err:
    IMPORT_ERROR = err


def fix_num(n):
    if n == "N/A":
        n = None
    return n


def parse_gpu(gpu, gid):
    mem = gpu["fb_memory_usage"]
    used = fix_num(mem["used"])
    total = fix_num(mem["total"])
    compute = fix_num(gpu["utilization"]["gpu_util"])
    if compute:
        compute /= 100
    return {
        "device": gid,
        "product": gpu["product_name"],
        "memory": {
            "used": used,
            "total": total,
        },
        "utilization": {
            "compute": compute,
            "memory": total and used and (used / total),
        },
        "temperature": fix_num(gpu["temperature"]["gpu_temp"]),
        "power": fix_num(gpu["power_readings"]["power_draw"]),
        "selection_variable": "CUDA_VISIBLE_DEVICES",
    }


def is_installed():
    return IMPORT_ERROR is None


class DeviceSMI:
    def __init__(self) -> None:
        if IMPORT_ERROR is not None:
            raise IMPORT_ERROR

        try:
            nvmlInit()
            self.nvsmi = nvidia_smi.getInstance()

        except NVMLError_LibraryNotFound as err:
            raise NotAvailable() from err

        except NVMLError_DriverNotLoaded as err:
            raise NotAvailable() from err

    @property
    def arch(self):
        return "cuda"

    @property
    def visible_devices():
        return os.environ.get("CUDA_VISIBLE_DEVICES", None)

    def get_gpus_info(self):
        to_query = [
            "gpu_name",
            "memory.free",
            "memory.used",
            "memory.total",
            "temperature.gpu",
            "utilization.gpu",
            "utilization.memory",
            "power.draw",
        ]
        results = self.nvsmi.DeviceQuery(",".join(to_query))

        if not results or "gpu" not in results:
            return {}

        gpus = results["gpu"]
        if not isinstance(gpus, list):
            gpus = [gpus]

        return {str(i): parse_gpu(g, i) for i, g in enumerate(gpus)}

    def close(self):
        pass



# We import the instruments lazily
_instruments = {
    "log": "from .log import log",
    "dash": "from .dash import dash",
    "gpu_monitor": "from .gpu import gpu_monitor",
    "rate": "from .metric import rate",
    "early_stop": "from .manage import early_stop",
}


def __getattr__(attr):
    if attr in _instruments:
        exec(_instruments[attr], globals())
        return globals()[attr]



import json
import os
import select
import subprocess
import time
from dataclasses import dataclass
from typing import Callable

from voir.smuggle import Decoder, MultimodalFile


@dataclass
class Stream:
    pipe: object
    info: dict
    deserializer: Callable = None


@dataclass
class LogEntry:
    event: str
    data: object
    pipe: str = None

    def get(self, item, default):
        return getattr(self, item, default)

    def dict(self):
        return dict(self.__dict__)

    def json(self):
        return json.dumps(self.__dict__)


def run(argv, info, timeout=None, constructor=None, env=os.environ, **options):
    mp = Multiplexer(timeout=timeout, constructor=constructor)
    mp.start(argv, info=info, env=env, **options)
    return mp


class Multiplexer:
    def __init__(self, timeout=0, constructor=None):
        self.processes = {}
        self.blocking = timeout is None
        self.timeout = timeout
        self.constructor = constructor or LogEntry
        self.buffer = []

    def start(self, argv, info, env=os.environ, use_stdout=False, **options):
        if use_stdout:
            proc = subprocess.Popen(
                argv,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env={**env, "DATA_FD": "1", "PYTHONUNBUFFERED": "1"},
                **options,
            )
            os.set_blocking(proc.stdout.fileno(), False)
            os.set_blocking(proc.stderr.fileno(), False)

            dec = Decoder(proc.stdout)
            mout = MultimodalFile(dec, "out", name=proc.stdout.name)
            mdat = MultimodalFile(dec, "data", name=proc.stdout.name)

            streams = [
                Stream(pipe=mout, info={"pipe": "stdout"}, deserializer=None),
                Stream(pipe=proc.stderr, info={"pipe": "stderr"}, deserializer=None),
                Stream(pipe=mdat, info={"pipe": "data"}, deserializer=json.loads),
            ]

        else:
            r, w = os.pipe()
            proc = subprocess.Popen(
                argv,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                pass_fds=[w],
                env={**env, "DATA_FD": str(w), "PYTHONUNBUFFERED": "1"},
                **options,
            )
            readdata = open(r, "r", buffering=1)
            os.set_blocking(proc.stdout.fileno(), False)
            os.set_blocking(proc.stderr.fileno(), False)
            os.set_blocking(r, False)

            streams = [
                Stream(pipe=proc.stdout, info={"pipe": "stdout"}, deserializer=None),
                Stream(pipe=proc.stderr, info={"pipe": "stderr"}, deserializer=None),
                Stream(pipe=readdata, info={"pipe": "data"}, deserializer=json.loads),
            ]

        self.add_process(
            proc=proc,
            argv=argv,
            info=info,
            streams=streams,
        )

        self.buffer.append(
            self.constructor(
                event="start",
                data={
                    "command": argv,
                    "time": time.time(),
                },
                **info,
            )
        )
        return proc

    def add_process(self, *, proc, info, argv, streams):
        self.processes[proc] = (streams, argv, info)

    def _process_line(self, line, s, pinfo):
        try:
            if isinstance(line, bytes):
                line = line.decode("utf8")
            if s.deserializer:
                try:
                    data = s.deserializer(line)
                    if "$event" in data:
                        yield self.constructor(
                            event=data.pop("$event"),
                            data=data.pop("$data", None),
                            **data,
                            **pinfo,
                            **s.info,
                        )
                    else:
                        yield self.constructor(
                            event="data",
                            data=data,
                            **pinfo,
                            **s.info,
                        )
                except Exception as e:
                    yield self.constructor(
                        event="format_error",
                        data={
                            "line": line,
                            "type": type(e).__name__,
                            "message": str(e),
                        },
                        **pinfo,
                        **s.info,
                    )
            else:
                yield self.constructor(event="line", data=line, **pinfo, **s.info)
        except UnicodeDecodeError:
            yield self.constructor(event="binary", data=line, **pinfo, **s.info)

    def __iter__(self):
        yield from self.buffer
        self.buffer.clear()

        while self.processes:
            still_alive = set()
            to_consult = {}
            for proc, (streams, _, info) in self.processes.items():
                for s in streams:
                    entries = to_consult.setdefault(s.pipe, [])
                    entries.append((s, proc, info))

            ready, _, _ = select.select(to_consult.keys(), [], [], self.timeout)

            for r in ready:
                while line := r.readline():
                    for s, proc, info in to_consult[r]:
                        yield from self._process_line(line, s, info)
                        still_alive.add(proc)

            for proc, (streams, argv, info) in list(self.processes.items()):
                if proc not in still_alive:
                    ret = proc.poll()
                    if ret is not None:
                        del self.processes[proc]
                        yield self.constructor(
                            event="end",
                            data={
                                "command": argv,
                                "time": time.time(),
                                "return_code": ret,
                            },
                            **info,
                        )

            if not self.blocking:  # pragma: no cover
                yield None



import os

from ...errors import NotAvailable

IMPORT_ERROR = None
try:
    import sys

    sys.path.append("/opt/rocm/libexec/rocm_smi/")
    import rsmiBindings as rsmi

except ImportError as err:
    IMPORT_ERROR = err


def rsmi_ret_ok(my_ret, device=None, metric=None, silent=False):
    """Returns true if RSMI call status is 0 (success)

    If status is not 0, error logs are written to the debug log and false is returned

    @param device: DRM device identifier
    @param my_ret: Return of RSMI call (rocm_smi_lib API)
    @param metric: Parameter of GPU currently being analyzed
    """
    if my_ret != rsmi.rsmi_status_t.RSMI_STATUS_SUCCESS:
        err_str = rsmi.c_char_p()
        rsmi.rocmsmi.rsmi_status_string(my_ret, rsmi.byref(err_str))

        returnString = ""
        if device is not None:
            returnString += "%s GPU[%s]:" % (my_ret, device)
        if metric is not None:
            returnString += " %s: " % (metric)
        returnString += "%s\t" % (err_str.value.decode())

        raise RuntimeError(returnString)
        return False
    return True


def is_driver_initialized():
    """Returns true if amdgpu is found in the list of initialized modules"""
    import subprocess

    driverInitialized = ""
    try:
        driverInitialized = str(
            subprocess.check_output(
                "cat /sys/module/amdgpu/initstate |grep live", shell=True
            )
        )
    except subprocess.CalledProcessError:
        pass
    if len(driverInitialized) > 0:
        return True
    return False


def initialize_rsmi():
    """initializes rocmsmi if the amdgpu driver is initialized"""
    # Check if amdgpu is initialized before initializing rsmi
    if is_driver_initialized() is True:
        ret_init = rsmi.rocmsmi.rsmi_init(0)
        if ret_init != 0:
            raise NotAvailable(
                "ROCm SMI returned %s (the expected value is 0)", ret_init
            )
    else:
        raise NotAvailable("Driver not initialized (amdgpu not found in modules)")


def list_devices():
    """Returns a list of GPU devices"""

    numberOfDevices = rsmi.c_uint32(0)
    ret = rsmi.rocmsmi.rsmi_num_monitor_devices(rsmi.byref(numberOfDevices))

    if rsmi_ret_ok(ret):
        deviceList = list(range(numberOfDevices.value))
        return deviceList
    else:
        exit(ret)


def get_gpu_use(device):
    """Return the current GPU usage as a percentage"""
    percent = rsmi.c_uint32()

    ret = rsmi.rocmsmi.rsmi_dev_busy_percent_get(device, rsmi.byref(percent))

    if rsmi_ret_ok(ret, device, "GPU Utilization "):
        return percent.value

    return -1


def get_mem_info(device, memType="vram"):
    """Return the specified memory usage for the specified device

    @param device: DRM device identifier
    @param type: [vram|vis_vram|gtt] Memory type to return
    """
    memType = memType.upper()
    if memType not in rsmi.memory_type_l:
        raise RuntimeError("Invalid memory type %s" % (memType))

    memoryUse = rsmi.c_uint64()
    memoryTot = rsmi.c_uint64()
    memUsed = None
    memTotal = None

    ret = rsmi.rocmsmi.rsmi_dev_memory_usage_get(
        device, rsmi.memory_type_l.index(memType), rsmi.byref(memoryUse)
    )
    if rsmi_ret_ok(ret, device, memType):
        memUsed = memoryUse.value

    ret = rsmi.rocmsmi.rsmi_dev_memory_total_get(
        device, rsmi.memory_type_l.index(memType), rsmi.byref(memoryTot)
    )
    if rsmi_ret_ok(ret, device, memType + " total"):
        memTotal = memoryTot.value
    return (memUsed, memTotal)


def get_temp(device, sensor):
    """Display the current temperature from a given device's sensor

    @param device: DRM device identifier
    @param sensor: Temperature sensor identifier
    """
    temp = rsmi.c_int64(0)
    metric = rsmi.rsmi_temperature_metric_t.RSMI_TEMP_CURRENT
    ret = rsmi.rocmsmi.rsmi_dev_temp_metric_get(
        rsmi.c_uint32(device),
        rsmi.temp_type_lst.index(sensor),
        metric,
        rsmi.byref(temp),
    )
    if rsmi_ret_ok(ret, device, sensor, True):
        return temp.value / 1000
    return "N/A"


def get_power(device):
    """Return the current power level of a given device

    @param device: DRM device identifier
    """
    power = rsmi.c_uint32()
    ret = rsmi.rocmsmi.rsmi_dev_power_ave_get(device, 0, rsmi.byref(power))
    if rsmi_ret_ok(ret, device, "power"):
        return power.value / 1000000
    return "N/A"


def get_product_name(device):
    series = rsmi.create_string_buffer(256)

    rsmi.rocmsmi.rsmi_dev_name_get(device, series, 256)

    series = series.value.decode()

    return series


def is_installed():
    return IMPORT_ERROR is None


class DeviceSMI:
    def __init__(self) -> None:
        if IMPORT_ERROR is not None:
            raise IMPORT_ERROR

        initialize_rsmi()
        self.devices = list_devices()

    def get_gpu_info(self, device):
        util = get_gpu_use(device)
        used, total = get_mem_info(device)
        temp = get_temp(device, "edge")
        power = get_power(device)

        return {
            "device": device,
            "product": get_product_name(device),
            "memory": {
                "used": used // (1024**2),
                "total": total // (1024**2),
            },
            "utilization": {
                "compute": float(util) / 100,
                "memory": used / total,
            },
            "temperature": temp,
            "power": power,
            "selection_variable": "ROCR_VISIBLE_DEVICES",
        }

    @property
    def arch(self):
        return "rocm"

    @property
    def visible_devices(self):
        return os.environ.get("ROCR_VISIBLE_DEVICES", None)

    def get_gpus_info(self):
        gpus = dict()
        for device in self.devices:
            gpus[device] = self.get_gpu_info(device)
        return gpus

    def close(self):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['voir', 'voir.instruments', 'voir.instruments.gpu']

package_data = \
{'': ['*']}

install_requires = \
['giving>=0.4.2,<0.5.0',
 'omegaconf>=2.3.0,<3.0.0',
 'ovld>=0.3.2,<0.4.0',
 'ptera>=1.4.1,<2.0.0',
 'pynvml>=11.5.0,<12.0.0',
 'rich>=13.3.2,<14.0.0']

entry_points = \
{'console_scripts': ['voir = voir.cli:main']}

setup_kwargs = {
    'name': 'voir',
    'version': '0.2.7',
    'description': 'Instrument, extend and visualize your programs',
    'long_description': 'None',
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import functools
import inspect
from functools import partial

from ovld import meta, ovld


@ovld
def gated(flag: str):  # noqa: F811
    return partial(gated, flag)


@ovld
def gated(flag: str, help: str):  # noqa: F811
    return partial(gated, flag, help=help)


@ovld
def gated(flag: str, instrument: meta(callable), help: str = None):  # noqa: F811
    dest = flag

    def run(ov):
        yield ov.phases.init
        ov.argparser.add_argument(flag, action="store_true", dest=dest, help=help)
        yield ov.phases.parse_args
        if getattr(ov.options, dest):
            ov.require(instrument)

    run.instrument = instrument
    return run


@ovld
def parametrized(option: str, type=None, help=None, default=None):  # noqa: F811
    return partial(parametrized, option, type=type, help=help, default=default)


@ovld
def parametrized(  # noqa: F811
    option: str, instrument: meta(callable), type=None, help=None, default=None
):
    def run(ov):
        yield ov.phases.init
        ov.argparser.add_argument(option, type=type, help=help, default=default)
        yield ov.phases.parse_args
        ov.require(instrument)

    return run


def instrument_definition(fn):
    def wrapped(*args, **kwargs):
        def instrument(ov):
            yield from fn(ov, *args, **kwargs)

        return instrument

    return wrapped


def configurable(fn):
    argspec = inspect.getfullargspec(fn)
    argname = argspec.args[1]
    ann = argspec.annotations[argname]

    @functools.wraps(fn)
    def wrapped(ov):
        yield ov.phases.init

        ov.argparser.add_from_model(argname, ann)

        yield ov.phases.parse_args
        yield from fn(ov, getattr(ov.options, argname))

    return wrapped



import os
import sys

import pkg_resources

from .overseer import Overseer
from .run import collect_instruments, find_voirfiles


def collect_contrib_instruments():
    results = []
    for entry_point in pkg_resources.iter_entry_points("voir.instrument"):
        results.append(entry_point.load())
    return results


def main(argv=None):
    sys.path.insert(0, os.path.abspath(os.curdir))

    vfs = os.environ.get("VOIRFILE", None)
    if vfs is None:
        vfs = find_voirfiles(".")
    else:
        vfs = vfs.split()

    instruments = collect_instruments(vfs)
    instruments.extend(collect_contrib_instruments())

    ov = Overseer(instruments=instruments, logfile=int(os.environ.get("DATA_FD", 3)))
    ov(sys.argv[1:] if argv is None else argv)



import operator
import sys
from functools import reduce
from pathlib import Path
from runpy import run_path

from ovld import ovld

module = type(operator)


def find_voirfiles(script_path):
    script_path = Path(script_path).expanduser().absolute()
    cur = script_path

    if cur.is_file():  # pragma: no cover
        # Currently not used
        cur = cur.parent
        vf = (cur / script_path.stem).with_suffix(".voirfile.py")
        results = [vf]
    else:
        results = []

    while cur != cur.parent:
        results.append(cur / "voirfile.py")
        cur = cur.parent

    return [str(pth) for pth in results if pth.exists()]


@ovld
def _to_instruments(self, value: list):  # noqa: F811
    return reduce(operator.add, map(self, value), [])


@ovld
def _to_instruments(self, value: dict):  # noqa: F811
    return reduce(operator.add, map(self, value.values()), [])


@ovld
def _to_instruments(self, value):  # noqa: F811
    return [value]


def _collect_instruments(voirfile, i):
    name = "__voir__" if i == 0 else f"__voir{i}__"
    md = module(name)
    md.__file__ = voirfile
    glb = run_path(voirfile, init_globals=vars(md), run_name=name)
    sys.modules[name] = md
    pfx = "instrument_"
    if "__instruments__" in glb:
        return _to_instruments(glb["__instruments__"])
    else:
        results = []
        for name, value in glb.items():
            if name.startswith(pfx):
                results.extend(_to_instruments(value))
        return results


def collect_instruments(voirfiles):
    return reduce(
        operator.add,
        [_collect_instruments(vf, i) for i, vf in enumerate(voirfiles)],
        [],
    )



"""Phase and generator-based system to run plugins."""

import heapq
import inspect
import threading
import time
from contextlib import contextmanager
from itertools import count
from queue import Empty, Queue

from giving import give, given

_gid = count()


class StopProgram(BaseException):
    """Raise from a handler/instrument to stop the program."""

    # Inherit from BaseException so that it isn't caught by "except Exception"
    # in user code. In that sense it is supposed to work a bit like
    # KeyboardInterrupt.


class OverseerAbort(BaseException):
    pass


class Phase:
    """Phase of a process.

    Attributes:
        name: Name of the phase.
        status: Phase status ("pending", "done" or "running")
        running: Whether the phase is running or not.
        done: Whether the phase is done or not.
        value: Result of the phase.
        exception: If the phase failed, contains the corresponding
            exception. If the phase succeeded, this is None.
    """

    def __init__(self, name, status="pending"):
        self.name = name
        self.status = status
        self.value = None
        self.exception = None

    @property
    def done(self):
        return self.status == "done"

    @property
    def running(self):
        return self.status == "running"

    def __call__(self, priority=0):
        return PhaseWithPriority(phase=self, priority=priority)


class PhaseWithPriority:
    """Associate a phase to wait for to a priority."""

    def __init__(self, phase, priority):
        self.phase = phase
        self.priority = priority


class PhaseSequence:
    """Sequence of phases, also works as a namespace."""

    def __init__(self, **phases):
        self._sequence = list(phases.values())
        self.__dict__.update(phases)
        self._current = 0

    def __iter__(self):
        return iter(self._sequence)


class PhaseRunner:
    """Organizes and runs phases.

    Arguments:
        phase_names: The names of the phases. The Phase objects are created
            automatically.
        args: Positional arguments to give to each handler.
        kwargs: Keyword arguments to give to each handler.
    """

    def __init__(self, phase_names, args=(), kwargs={}):
        phases = {phase_name: Phase(phase_name) for phase_name in phase_names}
        self.phases = PhaseSequence(_boot=Phase("_boot", status="done"), **phases)
        self.phases.IMMEDIATE = self.phases._boot(priority=0)
        self.handlers = set()
        # The plan maps each phase to a heap queue. The heap queue contains
        # (-priority, gid, generator, requested_phase) tuples, where gid is a
        # monotonically increasing number used to ensure that we follow
        # require() order when priorities are equal. The requested_phase is what
        # the generator last yielded and determines what is sent or thrown to
        # it.
        self.plan = {phase: [] for phase in self.phases}
        self.handler_args = args
        self.handler_kwargs = kwargs
        self.status = "init"
        self._to_require = []

    def _require(self, func):
        """Add a new handler.

        The same ``func`` will only be added once.

        The callable will be called with ``self.handler_args`` and
        ``self.handler_kwargs``. If it returns a generator, the generator must
        yield phases from ``self.phases``. The generator is immediately executed
        for all phases that are already done, and then queued for the next phase
        that is either currently processed or to be processed in the future.

        Any errors in the handler are passed to ``self.on_error``.

        Arguments:
            func: A callable.
        """
        state = getattr(func, "__state__", func)

        if self.status == "init":
            self._to_require.append(func)
            return state

        if func in self.handlers:
            return state

        self.handlers.add(func)

        try:
            gen = func(*self.handler_args, **self.handler_kwargs)
        except StopProgram as stp:
            self.on_stop(*stp.args)
            raise
        except BaseException as exc:
            self.on_overseer_error(exc)
            return

        if not inspect.isgenerator(gen):
            return state

        self._step((0, next(_gid), gen, self.phases._boot))
        return state

    def require(self, *funcs):
        states = [self._require(func) for func in funcs]
        return states[0] if len(states) == 1 else states

    def stop(self, value=None):
        raise StopProgram(value)

    def abort(self, exc):
        raise OverseerAbort(exc)

    def on_overseer_error(self, e):
        pass

    def on_stop(self, value):
        self.status = "stopped"
        for entries in self.plan.values():
            for _, __, gen, ___ in entries:
                try:
                    gen.throw(StopProgram(value))
                except (StopProgram, StopIteration):
                    pass
                except BaseException as exc:
                    self.on_overseer_error(exc)
                    pass

    def _step(self, entry):
        """Step for one generator.

        Arguments:
            entry: A (priority, gid, generator, requested_phase) tuple.
        """
        _, gid, gen, next_phase = entry
        while True:
            next_phase, next_priority = self._step_one(gen, next_phase)
            if next_phase is None:
                return
            elif not next_phase.done:
                break
        heapq.heappush(self.plan[next_phase], (-next_priority, gid, gen, next_phase))

    def _step_one(self, gen, ph):
        """Run one step of the generator using the given phase.

        The generator is sent ph.value, or thrown ph.exception, depending on
        whether ph.exception is None or not. Any errors are caught and sent
        to the error handler.
        """
        try:
            if ph.exception is not None:
                try:
                    next_phase = gen.throw(ph.exception)
                except BaseException as exc:
                    if exc is not ph.exception:
                        # Note: StopIteration will follow this path
                        raise
                    else:
                        return None, None
            else:
                next_phase = gen.send(ph.value)
            if isinstance(next_phase, PhaseWithPriority):
                next_priority = next_phase.priority
                next_phase = next_phase.phase
            else:
                next_priority = 0
            if next_phase not in self.phases:
                raise Exception("Generator must yield a valid phase")
        except StopIteration:
            return None, None
        except StopProgram:
            raise
        except OverseerAbort as exc:
            raise exc
        except BaseException as exc:
            self.on_overseer_error(exc)
            return None, None
        return next_phase, next_priority

    @contextmanager
    def run_phase(self, phase):
        """Run a phase.

        Arguments:
            phase: One of the Phases in ``self.phases``.
        """
        result = exception = None

        def _set_value(value):
            nonlocal result
            result = value

        try:
            yield _set_value
        except BaseException as exc:
            exception = exc

        phase.status = "running"
        phase.value = result
        phase.exception = exception
        entries = self.plan[phase]

        try:
            while entries:
                # Note: existing coroutines can call require() to add new entries,
                # so the heap can become larger from an iteration to the next.
                entry = heapq.heappop(entries)
                self._step(entry)
            phase.status = "done"
        except OverseerAbort as exc:
            raise exc.args[0]
        else:
            if exception:
                raise exception

    def __call__(self, *args, **kwargs):
        if self.status != "init":
            raise Exception("Can only enter runner when status == 'init'")
        self.status = "running"
        try:
            for req in self._to_require:
                self.require(req)
            self.run(*args, **kwargs)
        except StopProgram as stp:
            self.on_stop(*stp.args)
            pass
        except BaseException:
            self.status = "error"
            raise
        else:
            self.status = "done"


class GivenPhaseRunner(PhaseRunner):
    """Phase runner that provides an interface to giving.give."""

    def __init__(self, phase_names, args=(), kwargs={}):
        super().__init__(
            phase_names=phase_names,
            args=args,
            kwargs=kwargs,
        )
        self._queue = Queue()
        self._thread = threading.current_thread()
        self._queue_called = False

    def give(self, **data):
        if threading.current_thread() is self._thread:
            give(**data)
        else:
            self.queue(**data)

    def _dump_queue(self):
        if self._queue_called:
            while True:
                try:
                    data = self._queue.get_nowait()
                    give(**data)
                except Empty:
                    break

    def queue(self, **data):
        """Give data into a queue, typically from other threads."""
        if not self._queue_called:
            qd = self.given.where("!$queued")

            @qd.subscribe
            def _(_):
                # Insert the queued data into the given() stream
                # whenever other data comes in
                self._dump_queue()

        self._queue_called = True

        data["$queued"] = time.time()
        self._queue.put(data)

    def __call__(self, *args, **kwargs):
        with given() as gv:
            self.given = gv
            super().__call__(*args, **kwargs)
            self._dump_queue()
